#!/usr/bin/perl
#/***************************************************************************
# *   Copyright (C) 2004 by Michael Moritz                                  *
# *   mimo@restoel.net                                                      *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU General Public License as published by  *
# *   the Free Software Foundation; either version 2 of the License, or     *
# *   (at your option) any later version.                                   *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU General Public License for more details.                          *
# *                                                                         *
# *   You should have received a copy of the GNU General Public License     *
# *   along with this program; if not, write to the                         *
# *   Free Software Foundation, Inc.,                                       *
# *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
# ***************************************************************************/
#
# gps-db-update.pl Database update script for gps
#
# update db for format used in gps version 0.93+
# use with version 1.x and higher if converting from 0.x db
#
# Note: by default this will just do a test run
#       debug info is written to STDOUT
#		errors go to STDERR
#		use gps-db-update.pl <config> 2> problems.txt to write
#			all problem reports to a file
#
#		This script can take very long to complete as
#		it reads every entry of your greylisting table
#		and does a reverse name lookup on the client ip
#		address.
#
use Socket;
use Fcntl;
use DBI;
use Sys::Syslog qw(:DEFAULT setlogsock);
#
# Usage: gps-db-update.pl <config file>
#
# change this in order to create the 
# triplet table with the new structure to 1
$create_table = 0;
# change this to 1 to write entries to the new table
# otherwise just a test run
$do_update = 0;
# change this to 1 to delete the old
# Triplets table after the update is finished
# I'd recommend doing this manually
# once you're sure everything went fine (drop table Triplets)
$delete = 0;
# set this to 0 to process the complete table
$limit = 10;
# end of configurables

sub usage {
	print "Usage: ".$0." configfile\n";
}

sub read_config {
	my $fn = shift;
	my $User_Preferences;
	open(CONFIG, $fn) or die "can't open $fn: $!";	
	while (<CONFIG>) {
	    chomp;                  # no newline
	    s/#.*//;                # no comments
	    s/^\s+//;               # no leading white
	    s/\s+$//;               # no trailing white
	    next unless length;     # anything left?
	    my ($var, $value) = split(/\s*=\s*/, $_, 2);
	    $User_Preferences->{$var} = $value;
	}
	close(CONFIG);
	return $User_Preferences;
} 

while ($option = shift(@ARGV)) {
	$configfile = $option;
}
if($configfile eq "") {
	&usage();
	exit 1;
}
print STDOUT "config file=".$configfile."\n";
my $hash_ref = &read_config($configfile);
#print "the keys... ", sort keys %$hash_ref, "...\n";

if($verbose) {
    while ( my ($key, $value) = each(%$hash_ref) ) {
        print "$key => $value\n";
    }
}
my $dbh = DBI->connect('DBI:'.$hash_ref->{dbtype}.':'.$hash_ref->{db_dbname}.':'.$hash_ref->{db_host},
                        $hash_ref->{db_username},
                        $hash_ref->{db_password} )
                or die "Couldn't connect to database: " . DBI->errstr;

print STDOUT "Connected to DB\n";
# begin
if($create_table) {
	$sql = "CREATE TABLE triplet (
		client_address VARCHAR(40),
		sender VARCHAR(160) NOT NULL,
		recipient VARCHAR(160) NOT NULL,
		ip64 NUMERIC(4,0) DEFAULT 0 NOT NULL,
		ip32 NUMERIC(4,0) DEFAULT 0 NOT NULL,
		ip16 NUMERIC(4,0) DEFAULT 0 NOT NULL,
		ip8 NUMERIC(4,0) DEFAULT 0 NOT NULL,
		count INTEGER NOT NULL DEFAULT '0',
		uts INTEGER NOT NULL,
		PRIMARY KEY (sender,recipient,ip64,ip32,ip16,ip8)
		)";
	print STDOUT $sql."\n";
	$dbh->do($sql)	or die "SQL: " . $dbh->errstr ."\n";
	
	$sql = "CREATE INDEX sender_recipient_index ON triplet(sender(15),recipient(15))";	
	print STDOUT $sql."\n";
	$dbh->do($sql)	or die "SQL: " . $dbh->errstr ."\n";
	
	#$sql = "CREATE INDEX client_ip_index on NewTrip ( ip64,ip32,ip16,ip8 )";
	#print STDOUT $sql."\n";
	#$dbh->do($sql)	or die "SQL: " . $dbh->errstr ."\n";
}
#main
$sql = 'SELECT * FROM Triplets '.($limit?"LIMIT $limit":"");
print STDOUT "$sql\n";
my $sth = $dbh->prepare( $sql ) or die "SQL: " . $dbh->errstr ."\n";
$sth->execute() or die "SQL: " . $dbh->errstr ."\n";
my $err_valid = 0;
my $err_resolve = 0;
my $err_short = 0;
my $err_insert = 0;
my $count = 0;
while( my $hash_ref = $sth->fetchrow_hashref() ) {
	$name = undef;
	if(!($hash_ref->{client_address} =~ /^[\d.]+$/)) {
		print STDERR "no a valid IP address: ".$hash_ref->{client_address}.", skipped\n";
		$err_valid++;
		while( my ($k, $v) = each %$hash_ref ) {
       			print STDERR "$k=$v ";
		}
		print STDERR "\n";
	} else {
		$addr = inet_aton($hash_ref->{client_address});
		$name = gethostbyaddr($addr, AF_INET);
		if(!$name) {
			print STDERR "failed to resolve: ".$hash_ref->{client_address}.",  skipped: ";
			while( my ($k, $v) = each %$hash_ref ) {
        			print STDERR "$k=$v ";
    			}
			print STDERR "\n";
			$err_resolve++;
		} else {
			#print split(/\./,$name)."\n";
			@arname = split(/\./,$name);
			$num = @arname;
			if($num > 1) {
				shift(@arname);
			}
			$name = lc(join('.',@arname));
		}
	}
	my ($ip64,$ip32,$ip16,$ip8) = split(/\./,$hash_ref->{client_address});
	$sql = "INSERT INTO triplet VALUES ( "
		.$dbh->quote($name).", "
		.$dbh->quote($hash_ref->{sender}).", "
		.$dbh->quote($hash_ref->{recipient}).", "
		."$ip64, $ip32, $ip16, $ip8, "
		.$hash_ref->{count}.", "
		.$hash_ref->{uts}." )"
		;
	print STDOUT $sql."\n";
	if($do_update) {
		        if(! $dbh->do($sql) )
                {
                        print STDERR "SQL: " . $dbh->errstr ."\n";
                        $err_insert++;
                }
	}
	$count++;
}	
$sth->finish;

if($delete) {
	$sql = "DROP TABLE Triplets";
	print STDOUT $sql."\n";
	$dbh->do($sql)	or die "SQL: " . $dbh->errstr ."\n";
}

print STDOUT "Processed $count records\n";
if($err_valid) {
	print STDERR "Error Stats: no valid IP address: $err_valid\n";
}
if($err_resolve) {
	print STDERR "Error Stats: failed to resolve: $err_resolve\n";
}
if($err_short) {
	print STDERR "Error Stats: host name too short: $err_short\n";
}
if($err_insert) {
	print STDERR "Error Stats: SQL insert errors: $err_insert\n";
}
#$sql = "DROP TABLE NewTrip";
#print STDOUT $sql."\n";
#$dbh->do($sql)
#	or die "SQL: " . $dbh->errstr ."\n";
$dbh->disconnect;
print STDOUT "Disconnected from DB\n";
exit 0;

